/*
 * kgraft_patch_bsc954005
 *
 * Fix for bsc#954005
 *
 *  SLE12 commit:
 *  6069ac158cd867cfc68fc29e9fa23db0a6a20475
 *
 *  Copyright (c) 2015 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/fs.h>
#include <linux/aio.h>

#include "local.h"

static int (*kgr_xfs_zero_eof)(struct xfs_inode	*ip, xfs_off_t offset, xfs_fsize_t isize);
static void (*kgr_xfs_ilock)(xfs_inode_t *, uint);
static void (*kgr_xfs_iunlock)(xfs_inode_t *, uint);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "xfs_zero_eof", (void *)&kgr_xfs_zero_eof },
	{ "xfs_ilock", (void *)&kgr_xfs_ilock },
	{ "xfs_iunlock", (void *)&kgr_xfs_iunlock },
};

static inline void kgr_xfs_rw_ilock(struct xfs_inode *ip, int type)
{
	if (type & XFS_IOLOCK_EXCL)
		mutex_lock(&VFS_I(ip)->i_mutex);
	kgr_xfs_ilock(ip, type);
}

static inline void kgr_xfs_rw_iunlock(struct xfs_inode *ip, int type)
{
	kgr_xfs_iunlock(ip, type);
	if (type & XFS_IOLOCK_EXCL)
		mutex_unlock(&VFS_I(ip)->i_mutex);
}

ssize_t kgr_xfs_file_aio_write_checks(struct kiocb *iocb, loff_t *pos,
	size_t *count, int *iolock, int *eventsent)
{
	struct file		*file = iocb->ki_filp;
	struct inode		*inode = file->f_mapping->host;
	struct xfs_inode	*ip = XFS_I(inode);
	int			error = 0;

restart:
	error = generic_write_checks2(iocb, pos, count, S_ISBLK(inode->i_mode));
	if (error) {
		return error;
	}

	if ((DM_EVENT_ENABLED(ip, DM_EVENT_WRITE) &&
	    !(file->f_mode & FMODE_NOCMTIME) && !*eventsent)) {
		int	dmflags = FILP_DELAY_FLAG(file);

		if (*iolock & XFS_IOLOCK_EXCL)
			 dmflags |= DM_FLAGS_IMUX; /* dmapi disable mutex flg */

		/*
		 * Fix bsc#954005
		 *
		 * -kgr_xfs_rw_iunlock(ip, XFS_ILOCK_EXCL);
		 */
		error = XFS_SEND_DATA(ip->i_mount, DM_EVENT_WRITE, ip,
				      *pos, *count, dmflags, iolock);
		if (error)
			return -error;
		/*
		 * Fix bsc#954005
		 *
		 * -kgr_xfs_rw_ilock(ip, XFS_ILOCK_EXCL);
		 */

	/* DMAPI NOSPACE will call this routine again. eventsent is retained
	 * to ensure that DM event is only sent once.
	 */
		*eventsent = 1;
		/*
		 * The iolock was dropped and reacquired in XFS_SEND_DATA
		 * so we have to recheck the size when appending.
		 * We will only "goto start;" once, since having sent the
		 * event prevents another call to XFS_SEND_DATA, which is
		 * what allows the size to change in the first place.
		 */
		if (kiocb_is_append(iocb) && *pos != XFS_ISIZE(ip))
			goto restart;
	}

	/*
	 * If the offset is beyond the size of the file, we need to zero any
	 * blocks that fall between the existing EOF and the start of this
	 * write.  If zeroing is needed and we are currently holding the
	 * iolock shared, we need to update it to exclusive which implies
	 * having to redo all checks before.
	 */
	if (*pos > i_size_read(inode)) {
		if (*iolock == XFS_IOLOCK_SHARED) {
			kgr_xfs_rw_iunlock(ip, *iolock);
			*iolock = XFS_IOLOCK_EXCL;
			kgr_xfs_rw_ilock(ip, *iolock);
			goto restart;
		}
		error = -kgr_xfs_zero_eof(ip, *pos, i_size_read(inode));
		if (error)
			return error;
	}

	/*
	 * Updating the timestamps will grab the ilock again from
	 * xfs_fs_dirty_inode, so we have to call it after dropping the
	 * lock above.  Eventually we should look into a way to avoid
	 * the pointless lock roundtrip.
	 */
	if (likely(!(file->f_mode & FMODE_NOCMTIME))) {
		error = file_update_time(file);
		if (error)
			return error;
	}

	/*
	 * If we're writing the file then make sure to clear the setuid and
	 * setgid bits if the process is not being run by root.  This keeps
	 * people from modifying setuid and setgid binaries.
	 */
	return file_remove_suid(file);
}

static int kgr_patch_bsc954005_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc954005_module_notify(struct notifier_block *nb,
		unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, "xfs"))
		return 0;

	ret = kgr_patch_bsc954005_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc954005_module_nb = {
	.notifier_call = kgr_patch_bsc954005_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc954005_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module("xfs")) {
		ret = kgr_patch_bsc954005_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc954005_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc954005_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc954005_module_nb);
}
